<?php

declare(strict_types=1);

namespace Drupal\role_enum;

use Drupal\Core\Config\ConfigEvents;
use Drupal\Core\Config\StorageTransformEvent;
use Drupal\role_enum\Attribute\AllPermissions;
use Drupal\role_enum\Attribute\Permissions;
use Drupal\role_enum\Attribute\Role;
use Drupal\role_enum\Enum\ConfigBackedRoleInterface;
use Drupal\role_enum\Enum\ExternalConfigBackedRoleInterface;
use Drupal\role_enum\Enum\PermissionsInterface;
use Drupal\user\PermissionHandlerInterface;
use Ramsey\Uuid\Uuid;
use Symfony\Component\DependencyInjection\Attribute\Autowire;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Service for transforming role enums to role configuration.
 */
final class ConfigTransformer implements EventSubscriberInterface {

  /**
   * Allowed character regex for derived role IDs.
   */
  public const CONFIG_ID_REGEX = '/[^a-z0-9_]+/';

  public const CONFIG_PREFIX = 'user.role.';

  public function __construct(
    #[Autowire(param: 'role_enum.uuid.namespace')]
    private string $uuidNamespace,
    private RoleMapping $mapping,
    private PermissionHandlerInterface $permissionHandler,
  ) {
  }

  public static function getSubscribedEvents(): array {
    $events = [];
    $events[ConfigEvents::STORAGE_TRANSFORM_IMPORT][] = ['onImportTransform'];
    $events[ConfigEvents::STORAGE_TRANSFORM_EXPORT][] = ['onExportTransform'];
    return $events;
  }

  public function onImportTransform(StorageTransformEvent $event): void {
    foreach ($this->generateCases() as [$r, $case]) {
      // UUID will be the same for enum given the same namespace.
      $uuid = Uuid::uuid5($this->uuidNamespace, \sprintf('%s::%s', $case::class, $case->name));

      /** @var array<\ReflectionAttribute<\Drupal\role_enum\Attribute\Role>> $attributes */
      $attributes = $r->getAttributes(Role::class);
      $definitionAttr = ($attributes[0] ?? NULL)?->newInstance();

      $hasAllPermissions = \count($r->getAttributes(AllPermissions::class)) > 0;
      $permissionStrings = ($hasAllPermissions === FALSE) ? $this->permissions($r) : [];

      $id = $this->mapping->getId($case);
      $definition = [
        'uuid' => (string) $uuid,
        'langcode' => 'en',
        'status' => TRUE,
        'dependencies' => [],
        'id' => $id,
        'label' => $definitionAttr?->label ?? \sprintf("🛂 %s::%s", $r->getDeclaringClass()->getShortName(), $case->name),
        'weight' => $definitionAttr?->weight ?? 0,
        'is_admin' => $hasAllPermissions,
        'permissions' => $permissionStrings,
      ];

      $event->getStorage()->write(\sprintf('%s%s', static::CONFIG_PREFIX, $id), $definition);
    }
  }

  public function onExportTransform(StorageTransformEvent $event): void {
    foreach ($this->generateCases() as [$r, $case]) {
      $id = $this->mapping->getId($case);
      // Config provided by Role Enum is essentially "ignored".
      $event->getStorage()->delete(\sprintf('%s%s', static::CONFIG_PREFIX, $id));
    }
  }

  /**
   * @phpstan-return \Generator<array{\ReflectionClassConstant, \Drupal\role_enum\Enum\RoleInterface}>
   */
  private function generateCases(): \Generator {
    foreach ($this->mapping->getEnumClasses() as $class) {
      /** @var \ReflectionClass<\Drupal\role_enum\Enum\RoleInterface> $reflectionClass */
      $reflectionClass = new \ReflectionClass($class);

      if ($reflectionClass->implementsInterface(ExternalConfigBackedRoleInterface::class)) {
        continue;
      }

      if (FALSE === $reflectionClass->implementsInterface(ConfigBackedRoleInterface::class)) {
        continue;
      }

      foreach ($class::cases() as $enum) {
        yield [new \ReflectionClassConstant($class, $enum->name), $enum];
      }
    }
  }

  private function permissions(\ReflectionClassConstant $r): array {
    /** @var array<string, array{origin: string}> $permissionStrings */
    $permissionStrings = [];

    /** @var array<\ReflectionAttribute<\Drupal\role_enum\Attribute\Permissions>> $attributes */
    $attributes = $r->getAttributes(Permissions::class);
    $permissionsAttr = ($attributes[0] ?? NULL)?->newInstance();
    foreach ($permissionsAttr?->permissions ?? [] as $permission) {
      if (\is_string($permission) && \class_exists($permission)) {
        $permissionEnum = new \ReflectionClass($permission);
        if ($permissionEnum->implementsInterface(PermissionsInterface::class)) {
          foreach ($permission::cases() as $permissionCase) {
            $permissionStrings[$permissionCase->id()] = ['origin' => $permissionCase::class . '::' . $permissionCase->name];
          }
        }
        else {
          throw new \Exception(\sprintf('class-string must be of %s, got: %s', PermissionsInterface::class, $permission));
        }
      }
      elseif (\is_string($permission)) {
        $permissionStrings[$permission] = ['origin' => \sprintf('string on %s::%s', $r->getDeclaringClass()->getName(), $r->getName())];
      }
      elseif ($permission instanceof PermissionsInterface) {
        $permissionStrings[$permission->id()] = ['origin' => $permission::class . '::' . $permission->name];
      }
      else {
        throw new \Exception('Unable to handle permission: ' . \var_export($permission, TRUE));
      }
    }

    if (TRUE === $permissionsAttr?->validate) {
      $differences = \array_diff_key($permissionStrings, $this->permissionHandler->getPermissions());
      if ($differences !== []) {
        $messages = [];
        foreach ($differences as $permissionName => ['origin' => $origin]) {
          $messages[] = \sprintf("`%s` provided by %s", $permissionName, $origin);
        }
        throw new \Exception(\sprintf("Permission validation failed for %s::%s, permissions do not exist: \n - %s", $r->getDeclaringClass()->getName(), $r->name, \implode("\n - ", $messages)));
      }
    }

    return \array_keys($permissionStrings);
  }

}
