<?php

declare(strict_types=1);

namespace Drupal\role_enum;

use Drupal\role_enum\Attribute\AsPermissions;
use Drupal\role_enum\Enum\PermissionsInterface;
use Symfony\Component\DependencyInjection\Compiler\CompilerPassInterface;
use Symfony\Component\DependencyInjection\ContainerBuilder;

/**
 * Role Enum compiler pass.
 *
 * Searches for permission enums in defined namespaces, then records the
 * classnames to the container, so they may be efficiently fetched at runtime.
 *
 * @internal
 */
final class RoleEnumPermissionCompilerPass implements CompilerPassInterface {

  public function process(ContainerBuilder $container): void {
    /** @var array<class-string, string> $containerNamespaces */
    $containerNamespaces = $container->getParameter('container.namespaces');

    /** @var array<string, array{pathname: string}> $containerModules */
    $containerModules = $container->getParameter('container.modules');

    $moduleDirs = \array_map(static function (array $definition) {
      return \dirname($definition['pathname']);
    }, $containerModules);
    // Longest directories first, to make array_find + str_contains correct.
    \uasort($moduleDirs, static function (string $a, string $b): int {
      return \strlen($b) <=> \strlen($a);
    });

    /** @var string[] $enumNamespaces */
    $enumNamespaces = $container->getParameter('role_enum.permissions.namespaces');

    $enumClasses = [];

    foreach ($this->getEnums($containerNamespaces, $enumNamespaces) as $r) {
      /** @var class-string<\Drupal\role_enum\Enum\PermissionsInterface> $objectListClassName */
      $moduleName = \array_find_key($moduleDirs, static function (string $moduleDirectory) use ($r) {
        // getFileName never false since its userland.
        // @phpstan-ignore-next-line
        return \str_contains($r->getFileName(), $moduleDirectory);
      });
      $objectListClassName = $r->getName();
      $enumClasses[$objectListClassName] = $moduleName;
    }

    $container->getDefinition(PermissionMapping::class)
      ->setArgument('$enumClasses', $enumClasses);
  }

  /**
   * Get enums for the provided namespaces.
   *
   * @param array<class-string, string> $namespaces
   *   An array of namespaces. Where keys are class strings and values are
   *   paths.
   * @param string[] $enumNamespaces
   *   Enum namespaces.
   *
   * @return \Generator<\ReflectionClass<\Drupal\role_enum\Enum\PermissionsInterface>>
   *   Generates class strings.
   *
   * @throws \ReflectionException
   */
  private function getEnums(array $namespaces, array $enumNamespaces): \Generator {
    foreach ($namespaces as $namespace => $dirs) {
      $dirs = (array) $dirs;
      foreach ($dirs as $dir) {
        foreach ($enumNamespaces as $enumNamespace) {
          $nsDir = $dir . '/' . \str_replace('\\', '/', $enumNamespace);
          if (!\file_exists($nsDir)) {
            continue;
          }
          $namespace .= '\\' . $enumNamespace;

          /** @var \RecursiveIteratorIterator<\RecursiveDirectoryIterator<\SplFileInfo>> $iterator */
          $iterator = new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($nsDir, \FilesystemIterator::KEY_AS_PATHNAME | \FilesystemIterator::CURRENT_AS_FILEINFO | \RecursiveDirectoryIterator::SKIP_DOTS), \RecursiveIteratorIterator::SELF_FIRST);
          foreach ($iterator as $fileinfo) {
            \assert($fileinfo instanceof \SplFileInfo);
            if ('php' !== $fileinfo->getExtension()) {
              continue;
            }

            /** @var \RecursiveDirectoryIterator|null $subDir */
            $subDir = $iterator->getSubIterator();
            if (NULL === $subDir) {
              continue;
            }

            $subDir = $subDir->getSubPath();
            $subDir = $subDir !== '' ? \str_replace(DIRECTORY_SEPARATOR, '\\', $subDir) . '\\' : '';

            /** @var class-string $class */
            $class = $namespace . '\\' . $subDir . $fileinfo->getBasename('.php');

            /** @var \ReflectionClass<\Drupal\role_enum\Enum\PermissionsInterface> $reflectionClass */
            $reflectionClass = new \ReflectionClass($class);
            if ($reflectionClass->isEnum() && $reflectionClass->implementsInterface(PermissionsInterface::class)) {
              if (\count($reflectionClass->getAttributes(AsPermissions::class)) > 0) {
                yield $reflectionClass;
              }
            }
          }
        }
      }
    }
  }

}
