<?php

declare(strict_types=1);

namespace Drupal\role_enum;

use Drupal\Core\Config\ConfigBase;
use Drupal\role_enum\Enum\ConfigBackedRoleInterface;
use Drupal\role_enum\Enum\ExternalConfigBackedRoleInterface;
use Drupal\role_enum\Enum\RoleInterface;
use Symfony\Component\DependencyInjection\Compiler\CompilerPassInterface;
use Symfony\Component\DependencyInjection\ContainerBuilder;
use function Safe\preg_replace as preg_replace;

/**
 * Role Enum compiler pass.
 *
 * Searches for enums in defined namespaces, then records the classnames
 * to the container, so they may be efficiently fetched at runtime.
 *
 * @internal
 */
final class RoleEnumRoleCompilerPass implements CompilerPassInterface {

  public function process(ContainerBuilder $container): void {
    /** @var array<class-string, string> $containerNamespaces */
    $containerNamespaces = $container->getParameter('container.namespaces');

    /** @var string[] $enumNamespaces */
    $enumNamespaces = $container->getParameter('role_enum.role.namespaces');

    $enumClasses = [];
    /** @var array<string, string> $enumIdMap */
    $enumIdMap = [];

    foreach ($this->getEnums($containerNamespaces, $enumNamespaces) as $r) {
      /** @var class-string<\Drupal\role_enum\Enum\RoleInterface> $objectListClassName */
      $objectListClassName = $r->getName();
      $enumClasses[] = $objectListClassName;

      if ($r->implementsInterface(ConfigBackedRoleInterface::class)) {
        foreach ($r->getReflectionConstants() as $constant) {
          $enumString = \sprintf('%s::%s', $objectListClassName, $constant->getName());
          /** @var \Drupal\role_enum\Enum\ConfigBackedRoleInterface $case */
          $case = \constant($enumString);
          $enumIdMap[$enumString] = $case instanceof ExternalConfigBackedRoleInterface
            ? $case->id()
            : \substr(
              // \Safe\preg_replace has bad hinting.
              // @phpstan-ignore-next-line
              preg_replace(ConfigTransformer::CONFIG_ID_REGEX, '', \strtolower(\sprintf('%s__%s', $r->getShortName(), $case->name))),
              // Offset from the end, ensuring it doesn't max out length plus
              // safety in case it gets converted to YAML later.
              offset: -(ConfigBase::MAX_NAME_LENGTH - \strlen(ConfigTransformer::CONFIG_PREFIX)),
            );
        }
      }
    }

    $container->getDefinition(RoleMapping::class)
      ->setArgument('$enumClasses', $enumClasses)
      ->setArgument('$enumIdMap', $enumIdMap);
  }

  /**
   * Get enums for the provided namespaces.
   *
   * @param array<class-string, string> $namespaces
   *   An array of namespaces. Where keys are class strings and values are
   *   paths.
   * @param string[] $enumNamespaces
   *   Enum namespaces.
   *
   * @return \Generator<\ReflectionClass<\Drupal\role_enum\Enum\RoleInterface>>
   *   Generates class strings.
   *
   * @throws \ReflectionException
   */
  private function getEnums(array $namespaces, array $enumNamespaces): \Generator {
    foreach ($namespaces as $namespace => $dirs) {
      $dirs = (array) $dirs;
      foreach ($dirs as $dir) {
        foreach ($enumNamespaces as $enumNamespace) {
          $nsDir = $dir . '/' . \str_replace('\\', '/', $enumNamespace);
          if (!\file_exists($nsDir)) {
            continue;
          }
          $namespace .= '\\' . $enumNamespace;

          /** @var \RecursiveIteratorIterator<\RecursiveDirectoryIterator<\SplFileInfo>> $iterator */
          $iterator = new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($nsDir, \FilesystemIterator::KEY_AS_PATHNAME | \FilesystemIterator::CURRENT_AS_FILEINFO | \RecursiveDirectoryIterator::SKIP_DOTS), \RecursiveIteratorIterator::SELF_FIRST);
          foreach ($iterator as $fileinfo) {
            \assert($fileinfo instanceof \SplFileInfo);
            if ('php' !== $fileinfo->getExtension()) {
              continue;
            }

            /** @var \RecursiveDirectoryIterator|null $subDir */
            $subDir = $iterator->getSubIterator();
            if (NULL === $subDir) {
              continue;
            }

            $subDir = $subDir->getSubPath();
            $subDir = $subDir !== '' ? \str_replace(DIRECTORY_SEPARATOR, '\\', $subDir) . '\\' : '';

            /** @var class-string $class */
            $class = $namespace . '\\' . $subDir . $fileinfo->getBasename('.php');

            /** @var \ReflectionClass<\Drupal\role_enum\Enum\RoleInterface> $reflectionClass */
            $reflectionClass = new \ReflectionClass($class);
            if ($reflectionClass->isEnum() && $reflectionClass->implementsInterface(RoleInterface::class)) {
              yield $reflectionClass;
            }
          }
        }
      }
    }
  }

}
