<?php

declare(strict_types=1);

namespace Drupal\Tests\role_enum\Kernel;

use Drupal\Core\Config\StorageCacheInterface;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\DependencyInjection\ContainerBuilder;
use Drupal\Core\Lock\NullLockBackend;
use Drupal\KernelTests\KernelTestBase;
use Drupal\role_enum_test\Entity\RoleEnumTestUser;
use Drupal\role_enum_test\Role\TestRoles;
use Drupal\role_enum_test\Role\TestVirtualRoles;
use Drupal\user\Entity\Role;
use Drupal\user\Entity\User;
use Drupal\user\PermissionHandlerInterface;

/**
 * Tests Role Enum property and trait.
 */
final class RoleEnumTest extends KernelTestBase {

  protected static $modules = [
    'user',
    'bca',
    'role_enum',
    'role_enum_test',
  ];

  protected function setUp(): void {
    parent::setUp();
    $this->installEntitySchema('user');
    Role::create(['id' => TestRoles::Foo->value, 'label' => 'Test'])->setSyncing(TRUE)->save();
    Role::create(['id' => TestRoles::Bar->value, 'label' => 'Test'])->setSyncing(TRUE)->save();
  }

  public function testRolePermissions(): void {
    // Configuration sync inspired by config_filter.
    /** @var \Drupal\Core\Config\StorageInterface $active */
    $active = $this->container->get(StorageCacheInterface::class);
    $this->copyConfig($this->getImportStorage(), $active);

    // Ensure the imported config sticks and is available.
    static::assertEqualsCanonicalizing([
      'testvirtualroles__role1',
      'testvirtualroles__role2',
    ], \array_keys(Role::loadMultiple()));

    $user1 = RoleEnumTestUser::create(['name' => $this->randomMachineName()]);
    $user1->r->add(TestVirtualRoles::Role1);
    $user1->save();

    $user2 = RoleEnumTestUser::create(['name' => $this->randomMachineName()]);
    $user2->r->add(TestVirtualRoles::Role2);
    $user2->save();

    // User 1 has no permissions granted by TestVirtualRoles::Role1.
    static::assertFalse($user1->hasPermission('role_enum_test one a'));
    static::assertFalse($user1->hasPermission('role_enum_test one b'));
    static::assertFalse($user1->hasPermission('role_enum_test two c'));
    static::assertFalse($user1->hasPermission('role_enum_test two d'));
    static::assertFalse($user1->hasPermission('role_enum_test three e'));
    // User 2 has permissions granted by TestVirtualRoles::Role1.
    static::assertTrue($user2->hasPermission('role_enum_test one a'));
    static::assertFalse($user2->hasPermission('role_enum_test one b'));
    static::assertTrue($user2->hasPermission('role_enum_test two c'));
    static::assertTrue($user2->hasPermission('role_enum_test two d'));
    static::assertTrue($user2->hasPermission('role_enum_test three e'));
  }

  public function testProperty(): void {
    $user = User::create(['name' => $this->randomMachineName()]);
    $user->save();
    static::assertInstanceOf(RoleEnumTestUser::class, $user);

    static::assertFalse($user->r->has(TestRoles::Foo));
    $user->r->add(TestRoles::Foo);
    static::assertTrue($user->r->has(TestRoles::Foo));
    $user->r->remove(TestRoles::Foo);
    static::assertFalse($user->r->has(TestRoles::Foo));

    // Chained.
    static::assertTrue($user->r->add(TestRoles::Foo)->r->has(TestRoles::Foo));
  }

  public function testTrait(): void {
    $user = User::create(['name' => $this->randomMachineName()]);
    $user->save();
    static::assertInstanceOf(RoleEnumTestUser::class, $user);

    static::assertFalse($user->roleEnumHas(TestRoles::Foo));
    $user->roleEnumAdd(TestRoles::Foo);
    static::assertTrue($user->roleEnumHas(TestRoles::Foo));
    $user->roleEnumRemove(TestRoles::Foo);
    static::assertFalse($user->roleEnumHas(TestRoles::Foo));

    // Chained.
    static::assertTrue($user->roleEnumAdd(TestRoles::Foo)->roleEnumHas(TestRoles::Foo));
  }

  public function testPermissions(): void {
    $permissions = \Drupal::service(PermissionHandlerInterface::class)->getPermissions();
    static::assertArrayHasKey('role_enum_test one a', $permissions);
    static::assertArrayHasKey('role_enum_test one b', $permissions);
    static::assertArrayHasKey('role_enum_test two c', $permissions);
    static::assertArrayHasKey('role_enum_test two d', $permissions);
    static::assertArrayHasKey('role_enum_test three e', $permissions);
    static::assertArrayHasKey('role_enum_test three f', $permissions);
    static::assertEquals('role_enum_test', $permissions['role_enum_test three e']['provider']);
    static::assertEquals('role_enum_test', $permissions['role_enum_test three f']['provider']);
  }

  /**
   * Trigger the import transformation and return its result.
   *
   * This triggers \Drupal\role_enum\ConfigTransformer::onImportTransform.
   */
  private function getImportStorage(): StorageInterface {
    // Transform the sync storage for import.
    return $this->container->get('config.import_transformer')->transform($this->container->get('config.storage.sync'));
  }

  public function register(ContainerBuilder $container): void {
    // Test users will get a uid1, make sure it doesn't affect asserts.
    $container->setParameter('security.enable_super_user', FALSE);
    // \Drupal\Core\Config\ImportStorageTransformer::transform always tries to
    // use lock, which is fragile.
    $container->register('lock', NullLockBackend::class);
  }

}
